;;; -*- Mode: Emacs-Lisp -*-

;;;; Common-code section for autoloading MUD mode.

;;; This file contains all of the modifications we'd like to make
;;; to the user's environment _except_ which actual client to load
;;; (so that the user can load the experimental, the released, or
;;; the obsolete version).  It is loaded by whichever toplevel
;;; loader file the user picked, and expects that the value of
;;; mud-client-directory-pathname has already been initialized.

;;; Record the version of this file.  This field is updated in
;;; some fashion every time this file changes.

(defconst mud-autoload-common-version "2")

;;; Set the user's load-path so the rest of the code gets loaded.
;;; Put our stuff _last_ in the load path so the user can override us
;;; completely if so desired.

(setq load-path
      (append load-path
	      (list (expand-file-name mud-client-directory-pathname))))

;;; Reset the "shipped" or default value of where to find the entry
;;; file so we find the common Media-Lab one instead, which will
;;; do a maybe-include of ~/.mud to allow user customization.

(setq mud-entry-file (concat mud-client-directory-pathname
			     "Media-Lab-mud-entry-file"))

;;;; Define some keys to make mudding easier.

;;; Note:  The variables defined below, by default, arrange to bind M-X mud
;;; to C-C C-M to make it easier to get to.  If you don't want this behavior,
;;; ensure that your init sets both variables setup-c-c-c-m-mud-binding and
;;; setup-other-c-c-mud-bindings to nil.

;;; Don't let the X Window System step on our keybindings!
;;;
;;; The easiest-to-type sequence is probably C-C C-M, which is theoretically
;;; reserved for users because it starts with C-C, but the X Window System for
;;; gnumacs steps on this key sequence by default (a leftover from release 10,
;;; but pointless in release 11, because the window system really uses C-X
;;; C-@).  Unfortunately, the way in which is does this makes this rather
;;; messy to fix:  the user's init file gets run, then command-line (defined
;;; in startup.el) runs and loads either a terminal- or window-system-specific
;;; file.  In the case of X, that means we load x-win.el, which requires x-mouse.el,
;;; which bashes the binding, all _after_ the user's init was run, which means
;;; it's hard to fix.  Fortunately, _after_ loading terminal- and window-system-
;;; specific init files, normal-top-level and command-line-1 both try to run
;;; the single function of zero arguments defined in term-setup-hook and
;;; window-setup-hook.  (They arrange to only run any given hook once, by
;;; setting the value of each of these variables to nil once the hook has
;;; been run.)  Hence, we stick a function on window-setup-hook that checks
;;; to see if we're running X, and, if so, undoes the bashed binding.  We
;;; don't just unconditionally undo the binding (either always, or if we
;;; already find the binding bashed), because it's conceivable (however
;;; unlikely) that we'll be running in some _non_-X window system, which
;;; might have its own reasons to bash that binding, in which case we'd
;;; better leave it alone.  Got all that?
;;;
;;; Note that if your init has some other reason to also be mucking around
;;; with window-setup-hook, then you should change either the function below
;;; or your function to call the other, since this hook is just a funcall
;;; in version 18, not a list of functions, so you've gotta make a toplevel
;;; function that calls the rest and stick that in the hook.  *sigh*
;;; Nota bene:  We don't check for the hook already being set here.

(defvar setup-c-c-c-m-mud-binding t
  "*If non-nil, whether to bind M-X mud to C-C C-M.")

(defvar setup-other-c-c-mud-bindings nil
  "If non-nil, whether to bind M-X mud to C-C m, C-C C-C m, and C-C C-C C-M.")

(defun maybe-undo-c-c-c-m-binding (new-target)
  (if (eq window-system 'x)
      (define-key global-map "\C-c\C-m" new-target)))

(defun fix-x-binding-for-mud ()
  (maybe-undo-c-c-c-m-binding 'mud))

(if setup-c-c-c-m-mud-binding
    (setq window-setup-hook 'fix-x-binding-for-mud))

;;; If this file wasn't somehow loaded by the user's init file, it won't
;;; set things up soon enough to get the hook established to override the
;;; X bindings.  However, if that's the case, then we're being run _after_
;;; any terminal- or window-system-specific initializations have happened
;;; (e.g., the user just decided to load this file during some editing
;;; session, rather than from the init), so we can just smash the X binding
;;; in the straightforward way.
;;;
;;; Note that we also need to run the form below if this file _was_ loaded
;;; by the user's init file, but the user doesn't happen to be using X
;;; (since, in that case, no window-system hooks will be run anyway).

(if setup-c-c-c-m-mud-binding
    (define-key global-map "\C-c\C-m" 'mud))

;;; The following code is now obsolete as of version 2 and later of this file,
;;; but is preserved for use by people who don't want to use the mechanism above
;;; and want to use the bindings below.
;;;
;;; The easiest-to-type sequence is probably C-C C-M, which is theoretically
;;; reserved for users because it starts with C-C, but the X Window System for
;;; gnumacs steps on this key sequence by default (a leftover from release 10,
;;; but pointless in release 11, because the window system really uses C-X
;;; C-@).  Until the code to fix that is installed here, we work around the
;;; problem by binding C-C m (harder to type), and also by binding C-C C-C m
;;; and C-C C-C C-M, avoid the whole problem.

(if setup-other-c-c-mud-bindings
    (progn
      (define-key global-map "\C-c\m" 'mud)
      (define-key global-map "\C-c\C-C\m" 'mud)
      (define-key global-map "\C-c\C-c\C-m" 'mud)))

;;; End of file.
