;; $Id: emime.el,v 1.19 1995/05/29 21:01:43 marc Exp marc $

(provide 'emime)

;;; See RFC 1521 for details on MIME formats.

;; configuration variables

(defvar emime-active-face 'highlight
  "*Face to be used for active regions of an emime buffer")

(defvar emime-boundary-face 'secondary-selection
  "*Face to be used for boundarys in a multipart emime buffer")

;; internal buffer-local state

(make-variable-buffer-local 'emime-mode)
(defvar emime-mode nil
  "buffer-local boolean non-nil if this is an emime minor mode buffer")

(make-variable-buffer-local 'emime-overlays)
(defvar emime-overlays nil
  "buffer-local list of current overlays in the buffer")

(make-variable-buffer-local 'emime-actions)
(defvar emime-actions nil
  "buffer-local list of current active regions in the buffer")

(make-variable-buffer-local 'emime-mouse-activate-point-mode-specific)
(defvar emime-mouse-activate-point-mode-specific
  "*function to call first when a region is clicked by the mouse")

;; "top-level" functions

(defun emime-mode (&optional arg)
  (interactive)
  "Turn emacs mime minor mode on or off"

  ;; minor-mode setup stuff

  (or (assq 'emime-mode minor-mode-alist)
      (setq minor-mode-alist (cons '(emime-mode " emime") minor-mode-alist)))

  (or (assq 'emime-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'emime-mode emime-minor-mode-map)
		  minor-mode-map-alist)))

  (if (setq emime-mode
	    (if (null arg) (not emime-mode)
	      (> (prefix-numeric-value arg) 0)))
      (emime-process-buffer)
    (emime-unprocess-buffer)))

(defun emime-process-buffer ()
  "Processes buffer as a MIME messages, adding overlays as markup\n\
information to make things easier to read"
  (emime-unprocess-buffer)
  (if (string-equal "1.0"
		    (or (emime-get-header-value "Mime-Version"
						(point-min) (point-max)) ""))
      (emime-process-region (point-min) (point-max))))

(defun emime-unprocess-buffer ()
  "Removes all MIME overlays from the buffer"
  (while emime-overlays
    (delete-overlay (car emime-overlays))
    (setq emime-overlays (cdr emime-overlays)))
  (setq emime-actions nil))

(defun emime-process-region (dmin dmax)
  (interactive "r")
  "Processes region as a MIME messages, adding overlays as markup\n\
information to make things easier to read"
  (save-excursion
    (goto-char dmin)
    (cond ((equal (char-after dmin) ?\n)
	   (forward-char))
	  ((not (search-forward "\n\n" dmax t))
	   (error "Can't find end of headers")))
    (apply 'emime-markup
	   (mapcar 'copy-marker (list dmin (point) (point) dmax)))))

(defun emime-markup (hmin hmax bmin bmax)
  (let* ((content-type
	  (emime-parse-content-type
	   (or (emime-get-header-value "Content-Type" hmin hmax)
	       "text/plain")))
	 (type (car content-type))
	 (subtype (car (cdr content-type)))
	 (parameters (car (cdr (cdr content-type)))))
    (save-excursion
      (set-buffer (marker-buffer bmin))
      (funcall (cond ((equal type "text")
		      'emime-markup-text)
		     ((equal type "multipart")
		      'emime-markup-multipart)
		     ((equal type "application")
		      'emime-markup-application)
		     ((equal type "message")
		      'emime-markup-message)
		     ((equal type "image")
		      'emime-markup-image)
		     ((equal type "audio")
		      'emime-markup-audio)
		     ((equal type "video")
		      'emime-markup-video))
	       hmin hmax bmin bmax type subtype parameters))))

;; header-processing functions

(defun emime-get-header-value (header hmin hmax)
  (save-excursion
    (if (markerp hmin)
	(emime-goto-marker hmin)
      (goto-char hmin))
    (let* ((begin-value-point
	    (search-forward-regexp (concat "^" header ":[ \t]*") hmax t))
	   (end-value-point
	    (search-forward-regexp "\n\\(\n\\|[^ \t]\\)" hmax 'foo))
	   (raw-header
	    (cond ((and begin-value-point (not end-value-point))
		   ;; XXX I'm not sure what to do if the headers end the
		   ;; XXX buffer, but this will do.
		   (buffer-substring begin-value-point (point)))
		  ((and begin-value-point end-value-point)
		   (buffer-substring begin-value-point
				     (- end-value-point 2))))))
      (if raw-header
	  (emime-canonicalize-header raw-header)))))

(defun emime-canonicalize-header (value)
  ;; unfold lines
  (while (string-match "\n[ \t]+" value)
    (setq value (concat (substring value 0 (match-beginning 0))
			" "
			(substring value (match-end 0)))))
  ;; remove parenthesized comments from non-quoted areas
  (let (left
	qstr
	(newval ""))
    (while (string-match "\"\\([^\"\\\\]\\|\\\\.\\)*\"" value)
      (setq left (substring value 0 (match-beginning 0)))
      (setq qstr (substring value (match-beginning 0) (match-end 0)))
      (setq value (substring value (match-end 0)))
      (while (string-match "[ \t]*(\\([^)]\\|\\\\.\\)*)[ \t]*" left)
	(setq left (concat (substring left 0 (match-beginning 0)) " "
			   (substring left (match-end 0)))))
      (setq newval (concat newval left qstr)))
    (while (string-match "[ \t]*(\\([^)]\\|\\\\.\\)*)[ \t]*" value)
      (setq value (concat (substring value 0 (match-beginning 0)) " "
			  (substring value (match-end 0)))))
    (setq value (concat newval value)))
  ;; remove leading whitespace
  (if (string-match "^[ \t]*" value)
    (setq value (substring value (match-end 0))))
    ;; remove trailing whitespace
  (if (string-match "[ \t]*$" value)
    (setq value (substring value 0 (match-beginning 0))))
  value)

(defmacro match-substring (str num)
  (` (let ((b (match-beginning (, num)))
	   (e (match-end (, num))))
       (if (and b e) (substring (, str) b e)))))

(defun emime-parse-content-type (value)
  (if (string-match "^\\([^/ \t]+\\)[ \t]*/[ \t]*\\([^/; \t]+\\)[ \t]*" value)
      (let ((type (downcase (match-substring value 1)))
	    (subtype (downcase (match-substring value 2)))
	    (parameters-string (substring value (match-end 0))))
	(let ((parameters nil))
	  (while (not (equal parameters-string ""))
	    (if (string-match
 		 "^;[ \t]*\\([^= \t]+\\)=\\(\\([^\"; \t]+\\)\\|\"\\(\\([^\"\\\\]\\|\\\\.\\)*\\)\"\\)[ \t]*"
		 parameters-string)
		(setq parameters
		      (cons
		       (cons (downcase (match-substring parameters-string 1))
			     (concat (match-substring parameters-string 3)
				     (match-substring parameters-string 4)))
		       parameters)
		      parameters-string
		      (substring parameters-string (match-end 0)))
	      (error "Content-Type header malformed")))
	  (list type subtype parameters)))))

(defun emime-goto-marker (m)
  (set-buffer (marker-buffer m))
  (goto-char (marker-position m)))

;; multipart support functions

;; returns ((preamble-begin) (preamble-end) (part-1-begin) (part-1-end) (part-2-begin) (part-2-end) ... (part-n-begin) (part-n-end) (postamble-begin) (postamble-end))

(defun emime-multipart-bounds (bmin bmax parameters)
  (let ((boundary (cdr (assoc "boundary" parameters)))
	;; start with the end of the {postamble, document}
	(bounds (list (marker-position bmax))))
    (save-excursion
      (emime-goto-marker bmin)
      (let ((blen (+ (length boundary) 4)))
	;; find last boundary
	;; no final \n because there might not be one at all.
	(or (search-forward (concat "\n--" boundary "--") bmax t)
	    (error "Can't find closing boundary"))
	;; prepend points before/after boundary, which are the end of the last
	;;   document part and the beginning of the postamble, respectively.
	(setq bounds (cons (point) bounds))
	(setq bounds (cons (- (point) (+ 1 blen)) bounds))
	;; for each boundary, 
	(while (search-backward (concat "\n--" boundary "\n") bmin t)
	  ;; prepend points before/after boundary, which are the end if the
	  ;;   preceding document part/preamble, and the beginning of the next
	  ;;   document part
	  (setq bounds (cons (+ (point) blen) bounds))
	  (setq bounds (cons (point) bounds)))
	;; once more, in case the first boundary is the first thing in the body
	(emime-goto-marker bmin)
	(if (looking-at (concat "--" boundary "\n"))
	  ;; prepend points before/after boundary, which are the end if the
	  ;;   preceding document part/preamble, and the beginning of the next
	  ;;   document part
	    (progn
	      (setq bounds (cons (+ (point) blen -1) bounds))
	      (setq bounds (cons (point) bounds))))
	;; prepend point of {beginning of preamble, end of headers}
	(cons (point) bounds)))))

(defun emime-last-part-iter (list1 list2)
  (if (null list1)
      (cons (car list2) (car (cdr list2)))
    (emime-last-part-iter (cdr list1) (cdr list2))))

(defun emime-last-part (bounds)
  (emime-last-part-iter (cdr (cdr (cdr (cdr bounds)))) bounds))

(defun emime-entity-max (bounds)
  (if (null (cdr bounds))
      (car bounds)
    (emime-entity-max (cdr bounds))))

;; overlay and active-region support

(defun emime-make-overlay (pmin pmax)
  (save-excursion
    (if (markerp pmin)
	(set-buffer (marker-buffer pmin)))
    (make-overlay pmin pmax)))

(defun emime-add-overlays (&rest overlays)
  (save-excursion
    (while overlays
      (overlay-put (car overlays) 'modification-hooks 
		   (list
		    (lambda (o pmin pmax)
		      ;; this is kind of a hack, but erase-buffer
		      ;; shouldn't fail.
		      (if (not (and (eq pmin (point-min))
				    (eq pmax (point-max))))
			  (error
			   "Modifications of MIME structure are not allowed")
			))))
      (set-buffer (overlay-buffer (car overlays)))
      (setq emime-overlays (cons (car overlays) emime-overlays))
      (setq overlays (cdr overlays)))))

(defun emime-add-active-region (bmin bmax action args)
  (save-excursion
    (set-buffer (marker-buffer bmin))
    (setq emime-actions (cons
			 (cons (cons (copy-marker bmin)
				     (copy-marker bmax))
			       (cons action args))
			 emime-actions))))

(defun emime-activate-point (point)
  (interactive "d")
  "If the current point is part of an active region, perform the action."
  (let* ((actions emime-actions))
    (while actions
      (if (and (>= point (car (car (car actions))))
	       (<= point (cdr (car (car actions)))))
	  (apply (car (cdr (car actions))) (cdr (cdr (car actions)))))
      (setq actions (cdr actions)))))

(defun emime-mouse-activate-point (click)
  (interactive "e") 
  "If the clicked point is part of an active region, perform the action."
  (or (and emime-mouse-activate-point-mode-specific
	   (funcall emime-mouse-activate-point-mode-specific
		  click))
      (let (p)
	(save-excursion
	  (mouse-set-point click)
	  (setq p (point)))
	(emime-activate-point p))))

(defun emime-activate-all-regions ()
  (interactive)
  "Activate all regions in the current buffer"
  (let* ((actions emime-actions))
    (while actions
      (apply (car (cdr (car actions))) (cdr (cdr (car actions))))
      (setq actions (cdr actions)))))

;; base64 decoding

(defconst emime-b64-decoding-vector
  (concat (make-string 8 ?\177)
	  "\177\377\377\177\177\377\177\177"
	  (make-string 16 ?\177)
	  "\377\177\177\177\177\177\177\177"
	  "\177\177\177\076\177\177\177\077"
	  "\064\065\066\067\070\071\072\073"
	  "\074\075\177\177\177\200\177\177"
	  "\177\000\001\002\003\004\005\006"
	  "\007\010\011\012\013\014\015\016"
	  "\017\020\021\022\023\024\025\026"
	  "\027\030\031\177\177\177\177\177"
	  "\177\032\033\034\035\036\037\040"
	  "\041\042\043\044\045\046\047\050"
	  "\051\052\053\054\055\056\057\060"
	  "\061\062\063"))

(defun emime-decode-b64-generic (do-output bmin bmax)
  (let ((temp 0)			; bitstring accumulator
	(count 0)			; quad counter for temp
	(three (make-string 1023 0))	; decoded bitstring
	(dindex 0)			; index into decoded bitstring
	(bstart bmin)			; index into encoded buffer
					;  corresponding to decoded bitstring
	(bindex bmin)			; index into encoded buffer
	str				; a block of encoded characters
	(strlen -1)			; number of characters in str
	(index 0)			; index into str
	dch)				; a single decoded character
    (while (< bindex bmax)
      (if (>= index strlen)
	  (setq strlen (min 154 (- bmax bindex))	; 154 is two lines
		str (buffer-substring bindex (+ bindex strlen))
		index 0))
      (setq dch (aref emime-b64-decoding-vector (aref str index))
	    index (1+ index)
	    bindex (1+ bindex))
      (cond ((eq dch ?\177)
	     (error "Bogus character in b64 encoding"))
	    ((eq dch ?\200)
	     (cond ((eq count 2)
		    (aset three dindex (lsh temp -4))
		    (setq dindex (+ 1 dindex)))
		   ((eq count 3)
		    (aset three dindex (lsh temp -10))
		    (aset three (+ 1 dindex) (lsh temp -2))
		    (setq dindex (+ 2 dindex))))
	     (setq bindex bmax))
	    ((eq dch ?\377))
	    (t
	     (setq temp (+ (* temp 64) dch)
		   count (1+ count))
	     (if (eq count 4)
		 (progn
		   (aset three dindex (lsh temp -16))
		   (aset three (+ 1 dindex) (lsh temp -8))
		   (aset three (+ 2 dindex) temp)
		   (setq dindex (+ 3 dindex))
		   (setq temp 0)
		   (setq count 0)
		   (if (eq dindex 1023)
		       (progn
			 (funcall do-output bstart bindex three)
			 (setq bstart bindex)
			 (setq dindex 0))))))))
    (if (not (eq dindex 0))
	(funcall do-output bstart bmax (substring three 0 dindex)))))

(defmacro emime-qp-hexchar-to-int (hexchar)
  (` (cond ((>= (, hexchar) ?a)
	    (+ 10 (- (, hexchar) ?a)))
	   ((>= (, hexchar) ?A)
	    (+ 10 (- (, hexchar) ?A)))
	   ((>= (, hexchar) ?0)
	    (- (, hexchar) ?0))
	   (t
	    (error "Quoted-Printable character is not valid")))))

(defun emime-decode-qp-generic (do-output bmin bmax)
  (save-excursion
    (save-restriction
      (narrow-to-region bmin bmax)
      (while (< (point) (point-max))
	(cond ((looking-at "\\([^=]*\\)[ \t]*\n")
	       (goto-char (match-end 0))
	       (funcall do-output (match-beginning 0) (match-end 0)
			(concat
			 (buffer-substring (match-beginning 1) (match-end 1))
			 "\n")))
	      ((looking-at "\\([^=]*\\)=")
	       (goto-char (match-end 1))
	       (if (< (match-beginning 1) (match-end 1))
		   (funcall do-output (match-beginning 1) (match-end 1)
			    (buffer-substring (match-beginning 1)
					      (match-end 1))))
	       (cond ((eq (char-after (+ (point) 1)) ?\n)
		      (forward-char 2)
		      (funcall do-output (- (point) 2) (point) ""))
		     (t
		      (forward-char 3)
		      (funcall do-output (- (point) 3) (point)
			       (+ (* 16 (emime-qp-hexchar-to-int
					 (char-after (- (point) 2))))
				  (emime-qp-hexchar-to-int
				   (char-after (- (point) 1))))))))
	      (t
	       (goto-char (point-max))
	       (funcall do-output (buffer-substring (point) (point-max)))))))))

(defun emime-decode-b64-to-process (process bmin bmax)
  (emime-decode-b64-generic
   (` (lambda (emin emax string)
	(process-send-string (, process) string)))
   bmin bmax))

(defun emime-decode-qp-to-process (process bmin bmax)
  (emime-decode-qp-generic
   (` (lambda (emin emax string)
	(process-send-string (, process) string)))
   bmin bmax))

(defun emime-decode-to-process (cte process bmin bmax)
  (save-excursion
    (set-buffer (marker-buffer bmin))
    (cond ((or (equal cte "7bit")
	       (equal cte "8bit")
	       (equal cte "binary"))
	   (process-send-region process bmin bmax))
	  ((equal cte "quoted-printable")
	   (emime-decode-qp-to-process process bmin bmax))
	  ((equal cte "base64")
	   (emime-decode-b64-to-process process bmin bmax)))))

;; active function targets

(defun emime-activate-external-body (hmin hmax parameters)
  (let ((access-type (cdr (assoc "access-type" parameters))))
    (let ((body-region
	   (save-excursion
	     (cond ((equal access-type "anon-ftp")
		    (let ((site (cdr (assoc "site" parameters)))
			  (directory (cdr (assoc "directory" parameters)))
			  (name (cdr (assoc "name" parameters))))
		      (find-file-other-window (concat "/ftp@" site ":/"
						      directory "/" name))))
		   ((or (equal access-type "local-file")
			(equal access-type "afs"))
		    (find-file-other-window (cdr (assoc "name" parameters)))))
	     (cons (copy-marker (point-min)) (copy-marker (point-max))))))
      (emime-markup hmin hmax (car body-region) (cdr body-region)))))

(defun emime-activate-image (bmin bmax cte)
  (let* ((process-connection-type nil)
	 ;; XXX should be parameterized.  Eventually, might want to
	 ;; parse .mailcap directly
	 (proc (start-process "foo" "*out*" "xv" "-")))
    (message (format "Decoding %s image..." cte))
    (emime-decode-to-process cte proc bmin bmax)
    (process-send-eof proc)
    (message (format "Decoding %s image... Done." cte))))

;; Each of these functions is called with hmin/hmax as the region containing
;; the headers, and bmin/bmax as the region containing the body of the
;; relevant entity.  Note that these may be markers in two different
;; buffers, as in the case of external-body documents

(defun emime-markup-text (hmin hmax bmin bmax type subtype parameters)
  )

(defun emime-markup-multipart (hmin hmax bmin bmax type subtype parameters)
  (let ((bounds (emime-multipart-bounds bmin bmax parameters)))
    (cond ((or (equal subtype "alternative")
	       (and (equal subtype "mixed")
		    (equal (length bounds) 6)))
	   ;; for now, always do the last one
	   (let* ((part (emime-last-part bounds))
		  ;; everything after the header, before the last part
		  (i1 (emime-make-overlay (car bounds) (car part)))
		  ;; everything after the last part
		  (i2 (emime-make-overlay (cdr part)
					  (emime-entity-max bounds))))
	     (overlay-put i1 'invisible 't)
	     (overlay-put i2 'invisible 't)
	     (emime-add-overlays i1 i2)
	     (emime-process-region (car part) (cdr part))))
	  ((equal subtype "mixed")
	   ;; make the preamble invisible
	   ;; bounds now starts with preamble-begin
	   (if (< (car bounds) (car (cdr bounds)))
	       (let ((preamble (emime-make-overlay (car bounds)
						   (car (cdr bounds)))))
		 (overlay-put preamble 'invisible t)
		 (emime-add-overlays preamble)))
	   (setq bounds (cdr bounds))
	   ;; bounds now starts with {preamble-end,boundary-begin}
	   ;; markup each boundary and part
	   (while (nth 4 bounds)
	     (let* ((boundary-begin (nth 0 bounds))
		    (part-begin (nth 1 bounds))
		    (part-end (nth 2 bounds))
		    b1 i1)
	       (if (eq (char-after boundary-begin) ?\n)
		   (setq boundary-begin (1+ boundary-begin)))
	       (setq
		b1 (emime-make-overlay boundary-begin part-begin)
		i1 (emime-make-overlay (+ 2 boundary-begin) (- part-begin 1)))
	       (overlay-put b1 'face emime-boundary-face)
	       (overlay-put i1 'invisible t)
	       (emime-add-overlays b1 i1)
	       (emime-process-region part-begin part-end)
	       (setq bounds (cdr (cdr bounds)))))
	   ;; markup the last boundary and make the postamble invisible
	   (let* ((boundary-begin (nth 0 bounds))
		  (preamble-begin (nth 1 bounds))
		  (preamble-end (nth 2 bounds))
		  (b1 (emime-make-overlay boundary-begin preamble-end))
		  (i1 (emime-make-overlay boundary-begin (+ 1 boundary-begin)))
		  (i2 (emime-make-overlay (+ 3 boundary-begin) (- preamble-begin 2)))
		  (i3 (emime-make-overlay (+ 1 preamble-begin) preamble-end)))
	     (overlay-put b1 'face emime-boundary-face)
	     (overlay-put i1 'invisible t)
	     (overlay-put i2 'invisible t)
	     (overlay-put i3 'invisible t)
	     (emime-add-overlays b1 i1 i2 i3))))))

(defun emime-markup-application (hmin hmax bmin bmax type subtype parameters)
  )

(defun emime-markup-message (hmin hmax bmin bmax type subtype parameters)
  (cond ((equal subtype "external-body")
	 (let ((i1 (emime-make-overlay hmin hmax))
	       (i2 (emime-make-overlay bmin bmax)))
	   (overlay-put i1 'invisible 't)
	   (overlay-put i2 'face emime-active-face)
	   (overlay-put i2 'mouse-face emime-active-face)
	   (emime-add-overlays i1 i2)
	   (emime-add-active-region bmin bmax
				    'emime-activate-external-body
				    (list bmin bmax parameters))))
	((equal subtype "rfc822")
	 (emime-process-region bmin bmax))))

(defun emime-markup-image (hmin hmax bmin bmax type subtype parameters)
  (let ((cte (emime-get-header-value "Content-Transfer-Encoding" hmin hmax))
	(i1 (emime-make-overlay hmin (- hmax 1)))
	(i2 (emime-make-overlay bmin bmax)))
    (if (not cte)
	(error "No Content-Transfer-Encoding header"))
    (overlay-put i1 'face emime-active-face)
    (overlay-put i1 'mouse-face emime-active-face)
    (overlay-put i2 'invisible 't)
    (emime-add-overlays i1 i2)
    (emime-add-active-region hmin (- hmax 1) 'emime-activate-image
			     (list bmin bmax (downcase cte)))))

(defun emime-markup-audio (hmin hmax bmin bmax type subtype parameters)
  )

(defun emime-markup-video (hmin hmax bmin bmax type subtype parameters)
  )

;; keys for emime buffers

(defvar emime-minor-mode-map (make-sparse-keymap)
  "keymap with emime-specific key bindings")

(define-key emime-minor-mode-map "\C-c\C-m\C-m" 'emime-activate-point)
(define-key emime-minor-mode-map [mouse-2] 'emime-mouse-activate-point)
