;;; -*- Mode: Emacs-Lisp -*-

;;;; Major Mode for talking to MUDs

;;; by James Aspnes (asp@cs.cmu.edu) and Stewart Clamen (clamen@cs.cmu.edu)
;;; and Pavel Curtis (pavel@parc.xerox.com), with still more modifications
;;; as noted in the changelog below.
;;;
;;; The original version of this came from
;;;    FTP.MATH.OKSTATE.EDU:/pub/muds/clients/mud.el
;;;
;;; 1989, 1990, 1991, 1992, 1993.
;;;
;;; MODIFICATION HISTORY
;;; 
;;; May/June 1990 - Generalized to handle SMUG and LPMUD
;;; 
;;; January  1991 - Added Pavel Curtis' MOO support and assorted bug fixes, 
;;;                 also changed process-status call to run under 18.56.
;;;
;;; February 1991 - Added input-history browsing support (M-p/M-n);
;;;                   commands for sending the latest item on the kill
;;;                   ring (C-c C-y), optionally with each line bracketed by a
;;;                   given prefix and suffix (C-u C-c C-y); and a command to
;;;                   clear the current input line (C-c C-u).
;;;                 Added support for adding/overriding server definitions
;;;                   and the default server in one's .mud file.
;;;                 Fixed some bugs.
;;;                 Added support for people who prefer that the type-in
;;;                   point for a MUD buffer stay glued to the bottom of the
;;;                   window (see the 'mud-use-entire-window' option).
;;;
;;; January 1993 - [Lenny Foner <foner@media.mit.edu>]
;;;                 Part 1 of a bunch of cleanups and extensions:
;;;                 . Rearranged user variables to put them near the front
;;;                   and easier to find.
;;;                 . Fixed a bunch of scoping problems that were causing the
;;;                   wrong filters to be run or momentarily caused the
;;;                   selection of nonexistent filters.
;;;                 . Added prompting in the echo area for connections, so the
;;;                   password doesn't show up in the transcript, including
;;;                   remembering passwords for later connections.
;;;                 . Added maybe-include feature for optional inclusion of
;;;                   entry files, and added loop-detection capability and
;;;                   lisp-style comments (semicolons as well as hashmarks).
;;;                 . Removed a big chunk of exactly-duplicated code that was
;;;                   clearly a mark & yank problem somewhere in the dim past.
;;;                 . Added a large number of comments about how the code
;;;                   works.
;;;                 . Half-fixed a bug with connections being left around if
;;;                   the buffer was killed (the real fix awaits either a
;;;                   version of kill-buffer that does what it claims to, or
;;;                   gnumacs version 19, which incorporates the necessary
;;;                   hook).
;;;                 . Added auto-filling of server output.
;;;                 . Added some verb/property editing functionality.
;;;                 . Added the ability to figure out what version of the
;;;                   client you're running, for reporting bugs.
;;;                 . Cleaned up some self-documentation.

;;;; Current version of this file.

;;; This field should be updated every time the publicly-accessable
;;; version of this file is modified, so I can track which version has
;;; which bugs.

(defvar mud-client-version "10")

;;; How the user can figure out what version is being run, to report bugs.

;;; We do boundp checks here in case this file was not loaded via the
;;; standard mechanism in the public directory.  This will let me know
;;; that this is the case, without totally breaking the ability of the
;;; user to tell me the running version.
(defun compute-mud-client-version ()
  (concat "MUD client version " mud-client-version
	  (if (boundp 'mud-autoload-branch-version)
	      (concat ", branch " mud-autoload-branch-version))
	  (if (boundp 'mud-autoload-common-version)
	      (concat ", common " mud-autoload-common-version))
	  "."))

(defun insert-mud-client-version ()
  "Insert the current MUD client version into the buffer at point."
  (interactive)
  (insert (compute-mud-client-version)))

(defun show-mud-client-version ()
  "Show the current version of the mud client being run."
  (interactive)
  (message (compute-mud-client-version)))

;;;; Definitions.

;;; +++
;;; Customization variables that are likely to be set in a user's init file.

(defvar mud-show-page nil
  "*If non-nil, pop up MUD buffer whenever a page arrives.")

(defvar mud-accept-reconnects nil
  "*If nil, reject reconnect signals. If non-nil, accept reconnect signals 
by breaking existing connection and establishing new connection.  If an
integer, spawn <n> connections before breaking any.")

(defvar mud-use-entire-window nil
  "*Try to keep the type-in point for a MUD buffer at the bottom of the window.")

(defvar mud-password-alist ()
  "Holds an alist of server names, character names, and passwords.
If you don't want this info remembered, set this to any non-list
value (e.g., 'foo).")

(defvar mud-macro-commands-alist (list (cons "nil" ""))
  "*Alist of macros (keyed by strings)")
(make-variable-buffer-local 'mud-macro-commands-alist)

;;; Set up filling for mud buffers.
;;; By default, the fill-column in a mud-mode buffer is the width of the entire screen.
;;; That's almost always too wide, so set it down to the minimum of the screen width
;;; and a more reasonable value.
(defvar mud-default-fill-column 78
  "*Maximum width to fill output in a MUD-mode buffer.
Note that your typin is _not_ wrapped to this width, since that would confuse the
logic that transmits the entire line you type to the server when you type Return.
This variable effects only output from the server, not your typein.")

;;; To turn off filling completely, redefine this function to do nothing.
(defun better-mud-filling ()
  (setq fill-column (min fill-column mud-default-fill-column)))

(setq mud-mode-hook 'better-mud-filling)
;;; ---

;;; +++
;;; The following two variables _might_ be useful to change in a user's init file,
;;; but probably not.

;;; This variable, in addition to what's read in the mud entry files, determines
;;; which muds we know about.  In general, users should use the mud entry files,
;;; since it's possible to update them without recompiling this code, at the
;;; small cost of parsing those files once per session.  However, this variable
;;; is useful for naming servers if you want to put an embedded space in the name.
;;; Users are urged to PUSH, not SETQ, this value, in their init files, just in
;;; case this code actually sets up some servers sometime in the future.  A
;;; commented-out example appears below.

(defconst mud-default-servers
  '(
;;; SERVER NAME		MUD		SITE				PORT
;   ("TinyMUD Classic"	TinyMUD		"fuzine.mt.cs.cmu.edu"		2323)
    ))

;;; Since the lab-wide entry file defines a default using "default-server",
;;; there's no point changing this unless you want a different default server,
;;; _and_ you don't want to do it in your own entry file.  This should be a
;;; string, e.g., "MediaMOO".
(defvar mud-default-default-server nil
  "Default 'default-server' name.")
;;; ---

;;; +++
;;; Everything below here _should not_ be set by the user in an init file.
;;; ---

;;; This variable is actually set in the Media-Lab version of this by the
;;; common-autoload code to something other than ~/.mud (at the moment,
;;; to "/usr/public/pub/MediaMOO/Clients/Emacs/Media-Lab-mud-entry-file").
;;; By default, that file also includes ~/.mud to allow user customization.
(defvar mud-entry-file "~/.mud"
  "*Pathname to location of MUD address/character/password file, which might include others.")

(defvar muds nil "List of all defined MUD types")

;;; NOTE CAREFULLY that DEFMUD's method of hanging properties off symbols
;;; means that failing to have the symbol in your dynamic scope (yes, dynamic,
;;; Virginia, because that's how gnumacs-lisp works) means that you'll get
;;; NIL when you might have thought you shouldn't.  This has obviously
;;; already led to some subtle bugs in this code.  This should probably
;;; be replaced with DEFSTRUCT with all deliberate speed.  --- Foner 26 Dec 92.
(defmacro defmud (mud prompt connect-filter connect-command
		      filters command-filters
		      startup-hook page-regexp)
  (list 'progn
	(list 'defvar mud nil)
	(list 'setplist (list 'quote mud) nil)
	(list 'put (list 'quote mud) ''prompt prompt)
	(list 'put (list 'quote mud) ''connect-filter connect-filter)
	(list 'put (list 'quote mud) ''connect-command connect-command)
	(list 'put (list 'quote mud) ''filters filters)
	(list 'put (list 'quote mud) ''command-filters command-filters)
	(list 'put (list 'quote mud) ''startup-hook startup-hook)
	(list 'put (list 'quote mud) ''page-regexp page-regexp)
	(list 'if (list 'not (list 'memq   (list 'quote mud) 'muds))
	      (list  'setq 'muds (list 'cons  (list 'quote mud) 'muds)))
	(list 'quote mud)))

(defun mud-prompt () (get mud 'prompt))
(defun mud-connect-filter () (get mud 'connect-filter))
(defun mud-connect-command () (get mud 'connect-command))
(defun mud-filters () (get mud 'filters))
(defun mud-command-filters () (get mud 'command-filters))
(defun mud-startup-hook () (get mud 'startup-hook))
(defun mud-page-regexp () (get mud 'page-regexp))

;;; Equivalent mud types
(defmacro defeqmud (mud2 mud1)		; &&& Changed from eqmud to defeqmud by Foner, 26 Dec 92, for clarity.
  (list 'progn
	(list
	 'setplist (list 'quote mud2) (list 'symbol-plist (list 'quote mud1)))
	(list 'if (list 'not (list 'memq   (list 'quote mud2) 'muds))
	      (list 'setq 'muds (list 'cons  (list 'quote mud2) 'muds)))))

;;;; Definitions for particular MUDs.

(defmud TinyMUD
  ?>					; prompt char
  'tinymud-connect-filter
  "connect"
  'tinymud-filter-hook
  'nil
  'tinymud-mode-hook
  "\\(You sense that [^ ]* is looking for you in \\|\\w+ pages: \\)"
 )

(defeqmud TinyMUCK TinyMUD)		; &&& Changed from eqmud to defeqmud by Foner, 26 Dec 92, for clarity.
(defeqmud TinyMUSH TinyMUD)		; &&& Changed from eqmud to defeqmud by Foner, 26 Dec 92, for clarity.
(defeqmud TeenyMUD TinyMUD)		; &&& Changed from eqmud to defeqmud by Foner, 26 Dec 92, for clarity.

(defmud SMUG
  ?=
  'nil
  ""
  'smug-filter-hook
  'smug-macro-command-filter-hook
  'smug-mode-hook
  "You sense that [^ ]* is looking for you in "
 )

(defmud LPMUD
  ?\ 					; prompt char
  nil
  ""
  nil
  nil
  'tinymud-mode-hook
  "You sense that [^ ]* is looking for you in "
 )

(defmud MOO
  ?>
  'tinymud-connect-filter
  "connect"
  'moo-filter-hook
  'nil
  'moo-mode-hook
  "\\(You sense that [^ ]* is looking for you in \\)"
  )

;;;; Handling certain asynchronous events (pages and reconnections).

(defvar mud-reconnect-regexp
  "#### Please reconnect to \\([^@]*\\)@\\([^ @]*\\) *\\(\\|([^ @]*)\\) port \\([0-9]+\\) ####.*$"
  "Regular expression for detecting reconnect signals.")

(defconst mud-new-connectionp nil
  "Flag to identify hail for new connection")

(defun mud-check-reconnect ()
  "Look for reconnect signal and open new connection if none to that
site already exists."
  (goto-char (point-min))
  (while (not (eobp))
    (if (and mud-accept-reconnects (looking-at mud-reconnect-regexp))
	(let ((mud-name (buffer-substring (match-beginning 1)
					  (match-end 1)))
	      (mud-server-addr (buffer-substring (match-beginning 2)
						 (match-end 2)))
	      (mud-server (and (not (eq (match-beginning 3)
					(match-end 3)))
			       (buffer-substring (1+ (match-beginning 3))
						 (1- (match-end 3)))))
	      (mud-port (string-to-int
			 (buffer-substring (match-beginning 4)
					   (match-end 4)))))
	  (delete-region (match-beginning 0) (match-end 0))
	  (let* ((mud-sys (assoc mud-name (mud-servers)))
		 (mud-buffer-name (concat "*" mud-name "*"))
		 (mud-buffer-process
		  (mud-find-existing-process mud-name)))

	    (cond
	     (mud-buffer-process	; Existing connection to that site...
	      (message "Connection to that site had already been established.")
	      (pop-to-buffer (process-buffer mud-buffer-process)))
	     ((not mud-server)
	      (message "GNU Emacs cannot handle nonsymbolic names.  Sorry."))
	     ((zerop mud-port)
	      (message "Illformed portal signal. Inform Builder."))
	     (t
	      (save-excursion
		(setq mud-new-connectionp mud-buffer-name)
		(open-mud mud-sys t)))))))
    (beginning-of-line 2)))

(defun mud-find-existing-process (name)
  "Find process of established Mud connection, if it exists"
  (let ((processes (process-list))
	(result nil))
    (while (and processes (not result))
      (if (string-equal (upcase (process-name (car processes)))
			(upcase name))
	  (setq result (car processes))
	(setq processes (cdr processes))))
    result))

(defun mud-check-page ()
  "Look for page message, and pop-up buffer if specified."
  (goto-char (point-min))
  (while (not (eobp))
    (if (and mud-show-page (looking-at (mud-page-regexp)))
	(progn
	  (display-buffer (current-buffer))
	  (message "You are being paged in %s"
		   (buffer-name (current-buffer)))))
    (beginning-of-line 2)))

;;;; Handling I/O to the MUD buffer:  filling, filtering, sending lines, etc.

(defun mud-fill-lines ()
  "Fill buffer line by line."
  (goto-char (point-min))
  (while (not (eobp))
    (let ((break (move-to-column (1+ fill-column))))
      (if (<= break fill-column)
	  (beginning-of-line 2)
	;; else fill
	(skip-chars-backward "^ \n")
	(if (bolp)
	    ;; can't fill, we lose
	    (beginning-of-line 2)
	  (delete-horizontal-space)
	  (insert ?\n))))))

(defvar mud-debug-packets nil
  "If non-nil, save protocol debugging information in a buffer.")

(defun mud-filter (proc string)
  "Filter for input from MUD process.  Calls MUD-specific filters as well. 
Also, if recently established new connection automatically, check to see 
if number of active connections exceeded connection limit and delete 
current process if so." 
  (save-excursion
    (if mud-debug-packets
	(progn
	  ;; Occasionally-useful debugging code.
	  (set-buffer (get-buffer-create "*MUD Packets*"))
	  (goto-char (point-max))
	  (insert "\n\n<<")
	  (insert string)
	  (insert ">>")))
    (set-buffer (process-buffer proc))
    (goto-char (marker-position (process-mark proc)))
    (let ((start (point)))
      (insert-before-markers string)
      (let ((end (point)))
	(goto-char start)
	(beginning-of-line nil)
	(save-restriction
	  (narrow-to-region (point) end)
	  (while (search-forward "\^m" nil t)
	    (replace-match ""))
	  (goto-char (point-min))
	  (run-hooks (mud-filters))))))
  (if (and (= scroll-step 1)
	   (eq (current-buffer) (process-buffer proc))
	   (= (point) (point-max)))
      (recenter -1))
  (if mud-new-connectionp
      (progn
	(if (or			     ; Do we close current connection?
	     (not (numberp mud-accept-reconnects))
	     (let ((c mud-accept-reconnects) (l (process-list)))
	       (while l
		 (if (and (eq (process-filter (car l)) 'mud-filter)
			  (memq (process-status (car l)) '(open run)))
		     (setq c (1- c)))
		 (setq l (cdr l)))
	       (< c 0)))
	    (progn
	      (delete-process (get-buffer-process (current-buffer)))
	      (kill-buffer (current-buffer))))
	(progn
	  (pop-to-buffer mud-new-connectionp)
	  (if (> (baud-rate) search-slow-speed) (recenter))
	  (setq mud-new-connectionp nil)))))

(defun mud-send ()
  "Send current line of input to a MUD."
  (interactive)
  (let ((proc (get-buffer-process (current-buffer))))
    (cond ((or (null proc)
	       (not (memq (process-status proc) '(open run))))
	   (message "Not connected---nothing sent.")
	   (insert ?\n))
	  (t
	   ;; process exists, send line
	   (let ((start (mud-find-input)))
	     (send-region proc start (point))
	     (send-string proc "\n")
	     (mud-remember-input (buffer-substring start (point)))
	     (goto-char (point-max))
	     (insert ?\n)
	     (move-marker (process-mark proc) (point))
	     (insert (mud-prompt))
	     (if (= scroll-step 1)
		 (recenter -1))
	     )))))

(defun mud-realign-and-send ()
  (interactive)
  (recenter 0)
  (mud-send))

(defun mud-cancel-input ()
  (interactive)
  (let ((start (mud-find-input)))
    (delete-region start (point))))

(defun mud-send-kill (arg)
  (interactive "P")
  (if arg
      (call-interactively 'mud-send-kill-prefix)
    (let ((proc (get-buffer-process (current-buffer))))
      (mud-send-string (car kill-ring) proc))))

(defun mud-send-kill-prefix (prefix suffix)
  (interactive "sPrefix: \nsSuffix: ")
  (let ((buf (current-buffer))
	(temp (generate-new-buffer " *MUD temp*")))
    (save-excursion
      (set-buffer temp)
      (yank)
      (let ((proc (get-buffer-process buf))
	    (case-replace nil))
	(goto-char (point-min))
	(untabify (point-min) (point-max))
	(while (re-search-forward "^\\(.*\\)$")
	  (replace-match (concat prefix "\\1" suffix)))
	(send-region proc (point-min) (point-max))
	(send-string proc "\n")		;; Flush remaining input
	(kill-buffer temp)))))

(defun mud-quit ()
  "Quit MUD process."
  (interactive)
  (if (yes-or-no-p "Are you sure you want to quit this MUD session? ")
      (delete-process (get-buffer-process (current-buffer)))))

;;;; MUD-mode itself.

;;; &&& Hmm.  Perhaps my alarmist comments about having to name symbols
;;; specially and having scoping problems were premature.  This looks
;;; like it gets set, and might be sufficient.  (But then why were certain
;;; things returning NIL without the binding?  Why was that LET there in
;;; OPEN-MUD?  Oohh---because the buffer variable MUD is set by MUD-INTERACTIVE-MODE,
;;; called by OPEN-MUD, but M-I-M gets called _AFTER_ the points in the code
;;; I was looking for, so there really were true scoping bugs, which still need
;;; to get fixed.  Perhaps the easiest way, again short of going to defstruct,
;;; would be to set the buffer variable MUD first thing, before we do _anything_
;;; else, and then we don't have to name arguments carefully or bind the symbol
;;; or do anything else weird to get the right behavior (and put a big comment
;;; at the setting explaining how important it is!).)  --- Foner 26 Dec 92
(defconst mud nil
  "Variable representing type of MUD active in current buffer")
(make-variable-buffer-local 'mud)

(defvar mud-mode-syntax-table nil
  "Syntax table used while in MUD mode.")

(defvar mud-interactive-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\n" 'mud-realign-and-send)
    (define-key map "\r" 'mud-send)
    (define-key map "\^c\^c" 'mud-quit)
    (define-key map "\^c\^m" 'mud-macro-command)
    (define-key map "\^c\^u" 'mud-cancel-input)
    (define-key map "\^c\^y" 'mud-send-kill)
    (define-key map "\ep" 'mud-previous-command)
    (define-key map "\en" 'mud-next-command)
    map)
  "Keymap for MUD interactive mode.")

(defun mud-interactive-mode (mud-type)	; &&& Added running of mud-mode-hook --- Foner 26 Dec 92.
  "Major Mode for talking to inferior MUD processes.

Commands: 
\\{mud-interactive-mode-map}
Global Variables: [default in brackets]

 mud-show-page					[nil]
    If non-nil, pop up MUD buffer whenever a page arrives.
 mud-accept-reconnects				[nil]
    If nil, reject reconnect signals. If non-nil, accept reconnect
    signals by breaking existing connection and establishing new
    connection.  If an integer, spawn that many connections before
    breaking any.
 mud-entry-file					[locally defined]
    Pathname to location of MUD address/character/password file.
 mud-use-entire-window				[nil]
    Try to keep the type-in point for the MUD buffer at the bottom
    of the window, so as not to have a half-window of blank space.

Hooks:
 mud-mode-hook runs for all MUD modes.
 In addition, each type of MUD gets its own hook run after mud-mode-hook,
   so for example, a MOO will get moo-mode-hook run after mud-mode-hook.

Defmud parameters:
 prompt
    Character to identify MUD command input.
 connect-filters
    Initial filter hooks (before login)
 filters
    List of hooks to call before displaying output from MUD
    process to MUD buffer.  [Default hooks support line-filling,
    page checking, and reconnect detection.]
 startup-hook
    Hook to run at startup.  Users wishing to use macros may want to
    bind it to the following in their .emacs file:

     (setq tinymud-mode-hook
           '(lambda ()
       	       (mud-load-macro-commands tinymud-macro-commands-file)))
"
  (interactive)
  (kill-all-local-variables)
  (setq mud mud-type)
  (setq mode-name (symbol-name mud-type))
  (setq major-mode 'mud-interactive-mode)
  (setq fill-column (1- (screen-width)))
  (setq indent-tabs-mode nil)
  (if (null mud-mode-syntax-table)
      (progn
	(setq mud-mode-syntax-table (make-syntax-table))
	(set-syntax-table mud-mode-syntax-table)
	(modify-syntax-entry ?_ "w")
	(modify-syntax-entry ?\[ "(]")
	(modify-syntax-entry ?\] ")["))
    (set-syntax-table mud-mode-syntax-table))
  (use-local-map (copy-keymap mud-interactive-mode-map))
  (make-local-variable 'mode-line-process)
  (let* ((s (and (concat "@" mud-server)))
	 (ss (cond ((not mud-accept-reconnects) "")
		   (t (if (> (length s) 20) (substring s 0 20) s)))))
    (setq mode-line-process (list (concat ss ":%s"))))
  (run-hooks 'mud-mode-hook (mud-startup-hook))) ; &&& Added running of mud-mode-hook --- Foner 26 Dec 92.

;;;; Initializing the connection to a MUD.

;;; There's some sort of bug in kill-buffer.  It claims to terminate
;;; subprocesses in the buffer, but in fact it does not for Mud processes,
;;; and I'm not sure why.  (It works correctly for the shell.)  So
;;; until I figure it out, at least don't let a zillion killed processes
;;; with open network connections sit around, by killing all the
;;; leftover processes the next time a user tries to connect somewhere.
;;; This should leave at most one of these hanging around at any one time.
;;; I suppose, if Emacs 19 doesn't fix this bug, at least it will define
;;; the kill-buffer-hook variable and I can clean things up there.
(defun mud (&optional autoconnect)	; Autoconnect might reasonably be nil, (4), or (16).
  "Connect to MUD, asking for the name of the site to which to connect.

With an optional argument, parse the entries in the file pointed to by
the value of the variable mud-entry-file for the character name and
password to connect with, and attempt a connection.  If there is no
contact info for the server in that file, prompts for and remembers a
character name and password for this server.  (To enable or disable
this cache, see the variable mud-password-alist.)

Note that, by including an argument and therefore having this command
prompt for a username and password, you can type them in without having
them ever appear (and be recorded in) the buffer which keeps a transcript
of your connection.  This makes typing your password more secure, because
it won't appear in the history:  if you type \"connect character password\"
directly, it will appear in the history like any other interaction with
the mud.  [A future version of this client won't even echo the password
as you type it in, making it even more secure from someone looking over
your shoulder.  The current version echoes it in the modeline only,
which is at least better than writing it into the transcript.]

As a special case, if the optional argument is C-U C-U, and if there
is no contact information for the given server in the entry file, we
reprompt for character name and password even if they were already
cached for this server.  This allows you to recover if you typo the
character name or password when prompted the first time.

If you'd like to have a character name supplied automatically but
don't want to put your password in an entry file, you can't do this
with current entry file format (which requires that a password be
specified if a character name is).  However, you _can_ do this by
putting the following in your ~/.emacs file:
    (setq mud-password-alist '((\"SomeMudName\" \"CharacterName\")))
...and invoking this command with a C-U C-U prefix argument the
first time you connect.  A future version of this client will
probably allow you to include a character name without a password
directly in an entry file, so you don't have to do this in your init."
  (interactive "P")
  (mud-cleanup-extra-processes)		; Do this first, for correct buffer naming.
  (let* ((choice
	  (assoc
	   (let ((completion-ignore-case t))
	     (completing-read "Server: "
			      (mud-servers)
			      nil t
			      (mud-default-server)))
	   (mud-servers)))
	 (mud-name (car choice))
	 (mud-sys (car (cdr choice)))
	 (mud-server (car (cdr (cdr choice))))
	 (mud-port (car (cdr (cdr (cdr choice))))))
    (open-mud mud-sys autoconnect)))

;;; Stolen from FTP.EL and modified.

;;; Checks to see if we already have a character name, or a character name
;;; and password, for HOST (some server).  HOST and CHARACTER are strings.
;;; If CHARACTER is NIL, we prompt for a character name, defaulting to
;;; the character name last used with this server, or the username of the
;;; logged-in user if we haven't connected to this server before.  After getting
;;; the name by argument or by prompt, we check the alist to see if we have
;;; a password remembered for the character.  If not, we prompt for it.
;;; Updates the alist and returns the character name and password.
;;; If NEW-PASSWORD-P is non-NIL, we always prompt for a new password,
;;; even if we already remember it, so the user can retype an incorrect one.
;;; NEW-PASSWORD-P doesn't affect prompting for the character name:  we'll
;;; prompt for one if unsupplied, else we'll use the one supplied.

(defun mud-default-password (host character)
  (if (not (listp mud-password-alist))
      ""
    (or (cdr (cdr (assoc host mud-password-alist)))
	(let ((l mud-password-alist))
	  (catch 'foo
	    (while l
	      (if (string= (car (cdr (car l))) character)
		  (throw 'foo (cdr (cdr (car l))))
		(setq l (cdr l))))
	    nil))
	"")))

(defun read-mud-character-and-password (host character new-password-p)
  (let (tem)
    (if (and (not new-password-p)
	     (listp mud-password-alist)
	     (setq tem (cdr (assoc host mud-password-alist)))
	     (or (null character)
		 (string= character (car tem))))
	tem
      (or character
	  (progn
	    (setq tem
		  (or (and (listp mud-password-alist)
			   (car (cdr (assoc host mud-password-alist))))
		      (user-login-name))) ; Best guess...
	    (setq character
		  (read-string (format "Character name for %s (default \"%s\"): "
				       host tem)))
	    (if (equal character "")
		(setq character tem))))
      (setq tem
	    (cons character
		  ;; This would be an excellent application for a non-echoing reader.
		  (read-string (format "Password for %s on %s: " character host)
			       (mud-default-password host character))))
      (message "")
      (if (and (listp mud-password-alist)
	       (not (string= (cdr tem) "")))
	  (setq mud-password-alist
		(cons (cons host tem)
		      mud-password-alist)))
      tem)))

;;; Utility for OPEN-MUD.
(defun mud-connect-from-entry (mud entry proc)
  ;; Expects a symbol naming a MUD, a two-element list of a
  ;; character name and a password, and the process into which
  ;; to stuff a string to do the connection.
  ;;
  ;; The argument name MUD is carefully chosen.  If it was not MUD,
  ;; there would have to be a LET here to bind that symbol anyway;
  ;; welcome to the wonderful world of dynamic scoping.  We need
  ;; to find MUD so that MUD-CONNECT-COMMAND sees it in its scope,
  ;; since that function just gets a plist entry off the symbol MUD.
  ;; (I suspect that this would all be clearer if rewritten
  ;; to use the DEFSTRUCT facility provided in CL.EL).
  ;; --- Foner 26 Dec 92.
  (mud-send-string
   (mapconcat '(lambda (x) x)
	      (cons
	       (mud-connect-command)
	       entry)
	      " ")
   proc))

;;; Utility for OPEN-MUD.
(defun mud-apply-filters (mud proc)
  ;; Slam in the correct filters before doing anything else.
  ;;
  ;; As above, the argument MUD is very carefully chosen.
  ;; Incidentally, the original code in OPEN-MUD suffered from
  ;; the apparent problem that MUD was never bound around
  ;; calls to MUD-CONNECT-FILTER!  This implies that that code
  ;; simply never worked anyway.  I wonder how many other subtle
  ;; bugs like this there are in this client, because dynamic
  ;; scoping and properties like this don't mix?  More reasons
  ;; why this should be rewritten to use DEFSTRUCT.
  (let ((filter (or (mud-connect-filter)
		    'mud-filter)))
    (set-process-filter proc filter)))

;;; &&& See comments at TINYMUD-CONNECT-FILTER about how it and OPEN-MUD
;;; don't seem to agree on who gets to set the postconnection (e.g., normal)
;;; filter---they both think they get to set it.  The only reason this ever
;;; worked is becuase OPEN-MUD got to set it last, and its setting was correct
;;; (TINYMUD-CONNECT-FILTER's idea was to set it to TINYMUD-FILTER, which
;;; nobody ever got around to writing, but that setting only lasted a millisecond
;;; before OPEN-MUD bashed it to MUD-FILTER, which is correct).  --- Foner 26 Dec 92
(defun open-mud (mud-sys autoconnect)	; Autoconnect might reasonably be nil, (4), or (16).
  (let ((index 0)
	(buf-name-root (concat "*" mud-name "*"))
	(buf-name nil))
    (while (and (get-buffer (setq buf-name
				  (if (= index 0)
				      buf-name-root
				    (format "%s<%d>" buf-name-root index))))
		(get-buffer-process buf-name)
		(process-status (get-buffer-process buf-name)))
      (setq index (+ index 1)))
    (let* ((buf (get-buffer-create buf-name))
	   (proc nil))
      ;; Note that we leave proc NIL here, and setq it at the last possible
      ;; moment to the network stream connected to the server.  In particular, we
      ;; do this after any prompting of the user might take place.  We do this
      ;; because the correct filters aren't in place until after the prompting, and
      ;; if we open the network connection before then, the server's welcome banner
      ;; gets printed using the wrong filter (on MOO at least, it comes out with
      ;; ^M's after each line).  Further, we can't simply open the network stream
      ;; and slam the filters into place immediately afterwards, because the existence
      ;; of the filters in place will screw up our attempts to prompt---the exact error
      ;; is that the mud-filter runs in a totally empty (mini-)buffer during prompting,
      ;; leading to a backtrace like this:
      ;;   Signalling: (wrong-type-argument integer-or-marker-p nil)
      ;;   goto-char(nil)
      ;;   (save-excursion ...)
      ;;   mud-filter(#<process MUD> "
\n")
      ;;   read-string("Character name for MediaMOO (default \"foner\"): ")
      ;; So instead we open the stream, and apply the filters, only once we're
      ;; sure we have the information to proceed.  Note that we leave proc NIL,
      ;; not unbound, to make it easier to check its value and connect outside the
      ;; autoconnection code if we're not autoconnecting.  --- Foner 26 Dec 92.
      (if autoconnect
	  ;; &&& Changed by Foner on 26 Dec 92 to check for a nonexistent entry and,
	  ;; if so, to prompt for a character name and password in the minibuffer.
	  ;; This is not only more correct (the old code simply transmitted a bogus
	  ;; "connect" command, eliciting a syntax error from the server), but also
	  ;; provides a way to type in a password that won't be left sitting in the
	  ;; transcript buffer.  (This would be the place for a nonechoing reader,
	  ;; if anyone was ever to write one.)  Note also that this code doesn't
	  ;; guarantee that we're actually connected with a character (e.g., the
	  ;; password could have been wrong), but at least it tries.
	  (let ((entry (mud-login-for-server mud-name)))
	    (if entry
		;; There's an entry here, so autoconnect using file information.
		(progn
		  (setq proc (open-network-stream "MUD" buf mud-server mud-port))
		  (mud-apply-filters mud-sys proc)
		  (mud-connect-from-entry mud-sys entry proc)))
	    ;; No entry here, so prompt for character name and password.
	    (let ((character-and-password
		   (read-mud-character-and-password
		    mud-name
		    nil
		    (equal autoconnect '(16))))) ; Reprompt even if cached, if called with C-U C-U.
	      ;; Must turn (character . password) into (character password), *sigh*.
	      (setq proc (open-network-stream "MUD" buf mud-server mud-port))
	      (mud-apply-filters mud-sys proc)
	      (mud-connect-from-entry mud-sys
				      (list (car character-and-password)
					    (cdr character-and-password))
				      proc))))
      ;; If we haven't autoconnected, connect now.
      (if (null proc)
	  (progn
	    (setq proc (open-network-stream "MUD" buf mud-server mud-port))
	    (mud-apply-filters mud-sys proc)))
      ;; Now go ahead and run the normal interaction code.
      (switch-to-buffer buf)
      (newline)
      (goto-char (point-max))
      (set-marker (process-mark proc) (point))
      (mud-interactive-mode mud-sys)
      (insert (mud-prompt))
      (cond (mud-use-entire-window
	     (make-local-variable 'scroll-step)
	     (setq scroll-step 1))
	    (t
	     (recenter '(4))))
      (mud-initialize-input-history))))

;;;; Input history maintenance.

(defun mud-make-history (size)
  ;; (head tail . vector)
  ;; head is the index of the most recent item in the history.
  ;; tail is the index one past the oldest item
  ;; if head == tail, the history is empty
  ;; all index arithmetic is mod the size of the vector
  (cons 0 (cons 0 (make-vector (+ size 1) nil))))

(defun mud-decr-mod (n m)
  (if (= n 0)
      (1- m)
    (1- n)))

(defun mud-history-insert (history element)
  (let* ((head (car history))
	 (tail (car (cdr history)))
	 (vec (cdr (cdr history)))
	 (size (length vec))
	 (new-head (mud-decr-mod head size)))
    (aset vec new-head element)
    (setcar history new-head)
    (if (= new-head tail)  ; history is full, so forget oldest element
	(setcar (cdr history) (mud-decr-mod tail size)))))

(defun mud-history-empty-p (history)
  (= (car history) (car (cdr history))))

(defun mud-history-ref (history index)
  (let* ((head (car history))
	 (tail (car (cdr history)))
	 (vec (cdr (cdr history)))
	 (size (if (<= head tail)
		   (- tail head)
		 (+ tail (- (length vec) head)))))
    (if (= size 0)
	(error "Ref of an empty history")
      (let ((i (% index size)))
	(if (< i 0)
	    (setq i (+ i size)))
	(aref vec (% (+ head i) (length vec)))))))

(defvar mud-input-history-size 20
  "The number of past input commands remembered for possible reuse")

(defvar mud-input-history nil)

(defvar mud-input-index 0)

(defun mud-initialize-input-history ()
  (make-local-variable 'mud-input-history)
  (make-local-variable 'mud-input-index)
  (setq mud-input-history (mud-make-history mud-input-history-size))
  (setq mud-input-index 0))

(defun mud-remember-input (string)
  (mud-history-insert mud-input-history string))

(defun mud-previous-command ()
  (interactive)
  (mud-browse-input-history 1))

(defun mud-next-command ()
  (interactive)
  (mud-browse-input-history -1))

(defun mud-browse-input-history (delta)
  (cond ((mud-history-empty-p mud-input-history)
	 (error "You haven't typed any commands yet!"))
	((eq last-command 'mud-browse-input-history)
	 (setq mud-input-index (+ mud-input-index delta)))
	(t
	 (setq mud-input-index 0)))
  (setq this-command 'mud-browse-input-history)
  (let ((start (mud-find-input)))
    (delete-region start (point))
    (insert (mud-history-ref mud-input-history mud-input-index))))

(defun mud-find-input ()
  (beginning-of-line 1)
  (let* ((proc (get-buffer-process (current-buffer)))
	 (start (max (process-mark proc) (point))))
    (if (equal (char-after start) (mud-prompt))
	(setq start (1+ start)))
    (goto-char start)
    (end-of-line 1)
    start))

;;;; Macros.

(defvar mud-current-process nil "Current MUD process")
(defvar mud-current-macro-commands-alist nil "Current MUD macro command alist")

(defvar mud-macro-expansion-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\^c\^c" 'mud-macro-send-and-destroy)
    (define-key map "\^c\^s" 'mud-macro-send)
    (define-key map "\^c\^]" 'mud-macro-abort)
    (define-key map "\^c\^d" 'mud-macro-define)
    map)
  "Keymap for mud-macro-expansion-mode.")

(defun mud-macro-expansion-mode ()
  "Major Mode for mucking with MUD macro expansion.
Commands:
\\{mud-macro-expansion-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (setq mode-name "MUD-Macro-Expansion")
  (setq major-mode 'mud-macro-expansion-mode)
  (setq indent-tabs-mode nil)
  (set-syntax-table mud-mode-syntax-table)
  (use-local-map mud-macro-expansion-mode-map)
  (make-local-variable 'mud-expansion-macro-name)
  (message "Use ^C^S to send, ^C^C to send and destroy, ^C^] to abort..."))

(defun mud-macro-define (name)
  "Define buffer as mud-macro."
  (interactive (list (completing-read "MUD Macro: "
				      mud-current-macro-commands-alist
				      nil nil
				      mud-expansion-macro-name)))
  (let ((oldval (assoc name mud-current-macro-commands-alist)))
    (if oldval
	(setcdr oldval (buffer-string))
      (setq 
       mud-current-macro-commands-alist
       (cons
	(cons name (buffer-string))
	mud-current-macro-commands-alist))))
  (if (y-or-n-p "Save to file? ")
      (mud-store-macro-commands
       (expand-file-name
	(read-file-name (concat "File to save to (default "
				mud-macro-commands-file
				"): ")
			"~/"
			mud-macro-commands-file)))))

(defun mud-macro-abort ()
  "Abort macro expansion buffer."
  (interactive)
  (kill-buffer (current-buffer))
  (delete-window))

(defun mud-macro-send ()
  "Send contents of macro expansion buffer."
  (interactive)
  (let ((str (buffer-string)))
    (mud-macro-send-2 str)))

(defun mud-macro-send-and-destroy ()
  "Send contents of macro expansion buffer and then kill the buffer."
  (interactive)
  (let ((str (buffer-string)))
    (mud-macro-abort)
    (mud-macro-send-2 str)))

(defun mud-macro-send-2 (str)
  (let ((proc mud-current-process))
    (set-buffer (process-buffer proc))
    (setq mud-macro-commands-alist mud-current-macro-commands-alist)
    (mud-send-string 
     (let ((start (point)))
       (insert str)
       (save-restriction
	 (narrow-to-region start (point))
	 (run-hooks (mud-command-filters))
	 (prog1
	     (buffer-string)
	   (delete-region (point-min) (point-max)))))
     proc)))

(defun mud-send-string (string proc)
  "Send STRING as input to PROC"
  (send-string proc (concat string "\n")))

(defun mud-load-macro-commands (filename)
  "Load file of mud-macros"
  (setq mud-macro-commands-alist
	(if (file-exists-p filename)
	    (progn
	      (setq mud-macro-commands-file filename)
	      (let ((tempbuf (get-buffer-create " *MUD Macros*"))
		    (buf (current-buffer)))
		(set-buffer tempbuf)
		(erase-buffer)
		(insert-file filename)
		(prog1 (car (read-from-string (buffer-string)))
		  (set-buffer buf))))
	  '("nil" . ""))))

(defun mud-store-macro-commands (filename)
  "Store MUD macros in filename"
  (interactive "FFile to save to: ")
  (setq mud-macro-commands-file filename)
  (save-excursion
    (let ((tmp (get-buffer-create " *Macros to write*")))
      (set-buffer tmp)

      (erase-buffer)
      (insert (prin1-to-string mud-current-macro-commands-alist))
      (write-file filename))))

(defun mud-macro-command (arg)
  "Insert into stream one of the commands in mud-macro-commands-alist.
Without command argument, opens buffer for editting.  With argument
sends alist entry directly to process."
  (interactive "P")
  (let ((macro
	 (assoc
	  (or (if (stringp arg) arg)
	      (completing-read "MUD Macro: "
			       mud-macro-commands-alist nil t nil))
	  mud-macro-commands-alist)))
    (let ((match (car macro))
	  (stuff (cdr macro)))
      (if (stringp stuff)
	  (let ((buff (get-buffer-create "*Expansion*"))
		(proc (get-buffer-process (current-buffer)))
		(alist mud-macro-commands-alist))
	    (if (not arg)
		(progn
		  (pop-to-buffer buff)
		  (erase-buffer)
		  (insert stuff)
		  (goto-char (point-min))
		  (mud-macro-expansion-mode)
		  (setq mud-expansion-macro-name match)
		  (setq mud-current-process proc)
		  (setq mud-current-macro-commands-alist alist)
		  )
	      (mud-send-string stuff proc)))))))

;;;; Reading from the entry file.

;;; +++
;;; *** FORMAT OF THE ENTRY FILE: ***
;;; server         <server-name>  <mud-type>  <host-name>  <port>
;;; default-server <server-name>
;;; include        <pathname>
;;; maybe-include  <pathname>
;;; <server-name>  <character-name>  <password>
;;; default        <character-name>  <password>
;;;
;;; Any number of spaces or tabs separate each field.
;;; Any line beginning with # or ; is a comment.
;;;
;;; The currently-known types of muds are:
;;;    TinyMUD (treated identically:  TinyMUCK, TinyMUSH, TeenyMUD)
;;;    SMUG
;;;    LPMUD
;;;    MOO
;;;
;;; The filename of an include-file will be merged with the directory
;;; of the file currently being read, which means that, unless you specify
;;; some directory as well as a name, the name will be in the same directory
;;; as the file currently being read.  Files may include each other arbitrarily
;;; deeply, and loops are detected and only traversed up to the first cycle
;;; (e.g., if A includes B includes A, we read A, then B, then stop, with
;;; no warning and no infinite loop).  If a file included with "include" does
;;; not exist, an error is signalled and no further lines are read.  If a file
;;; included with "maybe-include" does not exist, we simply ignore the attempted
;;; inclusion and continue.  This mechanism is used to include "~/.mud", which
;;; the user may modify to add privately-known muds or character names, by creating
;;; a file called "~/.mud" and adding data there, rather than in this file.
;;;
;;; File data is cached and only reread if the write-date on the file changes.
;;;
;;; Note that the format of this file mandates that server names do not have
;;; embedded spaces.  If you want to create server names with embedded spaces
;;; (e.g., "TinyMUD Classic" instead of "TinyMUD-Classic"), you'll have to use
;;; the variable mud-default-servers (see the source for how to do this).
;;; ---

(defvar mud-servers nil)
(defvar mud-default-server nil)
(defvar mud-logins nil)
(defvar mud-default-login nil)

(defvar mud-entry-file-dates nil)

(defun mud-match-field (i)
  (buffer-substring (match-beginning i) (match-end i)))

(defun mud-report-syntax-error ()
  (let ((start (point)))
    (end-of-line)
    (error (concat "Syntax error in MUD entry file " file ": "
		   (buffer-substring start (point))))))

(defun mud-file-directory (name)
  (let ((i (1- (length name))))
    (while (not (= (aref name i) ?/))
      (setq i (1- i)))
    (substring name 0 (1+ i))))

(defun mud-file-write-date (file)
  (nth 5 (file-attributes file)))

(defun mud-entry-pattern (keyword nargs)
  (let ((pattern "?$"))
    (while (> nargs 0)
      (setq pattern (concat "\\([^ \n]*\\) " pattern)
	    nargs (1- nargs)))
    (if (null keyword)
	pattern
      (concat keyword " " pattern))))

(defun mud-handle-include-file (file nonexistent-okay)
  ;; Canonicalize the included name for loop detection, since we
  ;; store the filenames in the file-date alist in canonical form also.
  (let ((included-file (expand-file-name
			(concat (mud-file-directory file)
				(mud-match-field 1)))))
    (if (not (assoc included-file mud-entry-file-dates))
	;; We should try to read the file, since we haven't read it before.
	;; However, if this is "include-optional", don't bitch if it's not there.
	(mud-parse-entry-file included-file nonexistent-okay))))

(defun mud-parse-entry-file (name &optional nonexistent-okay)
  (catch 'done
    (let ((file (expand-file-name name))
	  (old-buffer (current-buffer))
	  (buffer (generate-new-buffer " *MUD temp*")))
      (if (not (file-exists-p file))
	  (if nonexistent-okay
	      (throw 'done nil)
	    (error (concat "Can't find MUD entry file " file))))
      ;; We set this _before_ attempting to parse the file so that, win or lose,
      ;; we won't try again until the file is changed.  This has two benefits:
      ;;  . A syntax error in the file will only prompt an error the first time.
      ;;    After that, we'll ignore the file until it's fixed.
      ;;  . We can use this alist to do loop-detection for included files.
      (setq mud-entry-file-dates
	    (cons (cons file (mud-file-write-date file))
		  mud-entry-file-dates))
      (unwind-protect
	  (progn
	    (set-buffer buffer)
	    (buffer-flush-undo buffer)
	    (insert-file-contents file)
	    ;; Don't lose if no final newline.
	    (goto-char (point-max))
	    (or (eq (preceding-char) ?\n)
		(newline))
	    (goto-char (point-min))
	    ;; handle "\\\n" continuation lines
	    (while (not (eobp))
	      (end-of-line)
	      (cond ((= (preceding-char) ?\\)
		     (delete-char -1)
		     (delete-char 1)
		     (insert ?\ )))
	      (forward-char 1))
	    ;; simplify whitespace handling
	    (goto-char (point-min))
	    (while (re-search-forward "^[ \t]+" nil t)
	      (replace-match ""))
	    (goto-char (point-min))
	    (while (re-search-forward "[ \t]+" nil t)
	      (replace-match " "))
	    (goto-char (point-min))
	    (while (not (eobp))
	      (cond ((or (eolp) (looking-at "#") (looking-at ";"))) ; Flush comments.
		    ((looking-at "server ")
		     (let (port
			   type)
		       (if (or (not (looking-at (mud-entry-pattern "server" 4)))
			       (= (setq port (string-to-int
					      (mud-match-field 4)))
				  0)
			       (not (memq (setq type
						(intern (mud-match-field 2)))
					  muds)))
			   (mud-report-syntax-error))
		       (setq mud-servers
			     (cons (list (mud-match-field 1)
					 type
					 (mud-match-field 3)
					 port)
				   mud-servers))))
		    ((looking-at "default-server ")
		     (if (not (looking-at (mud-entry-pattern "default-server"
							     1)))
			 (mud-report-syntax-error))
		     (if (null mud-default-server)
			 (setq mud-default-server (mud-match-field 1))))
		    ((looking-at "include ")
		     (if (not (looking-at (mud-entry-pattern "include" 1)))
			 (mud-report-syntax-error))
		     (mud-handle-include-file file nil))
		    ((looking-at "maybe-include ")
		     (if (not (looking-at (mud-entry-pattern "maybe-include" 1)))
			 (mud-report-syntax-error))
		     (mud-handle-include-file file t))
		    ((looking-at "default ")
		     (if (not (looking-at (mud-entry-pattern "default" 2)))
			 (mud-report-syntax-error))
		     (if (null mud-default-login)
			 (setq mud-default-login (list (mud-match-field 1)
						       (mud-match-field 2)))))
		    ((looking-at (mud-entry-pattern nil 3))
		     (setq mud-logins (cons (list (mud-match-field 1)
						  (mud-match-field 2)
						  (mud-match-field 3))
					    mud-logins)))
		    (t (mud-report-syntax-error)))
	      (beginning-of-line 2)))
	(kill-buffer buffer)
	(set-buffer old-buffer)))))

(defun mud-check-entry-file ()
  (if (or (null mud-entry-file-dates)
	  (let ((dates mud-entry-file-dates))
	    (while (and dates
			(equal (cdr (car dates))
			       (mud-file-write-date (car (car dates)))))
	      (setq dates (cdr dates)))
	    (not (null dates))))
      (progn
	(setq mud-servers nil
	      mud-default-server nil
	      mud-logins nil
	      mud-default-login nil
	      mud-entry-file-dates nil)
	(if (file-exists-p mud-entry-file)
	    (mud-parse-entry-file mud-entry-file))
	(setq mud-servers (append (reverse mud-servers) mud-default-servers))
	(if (null mud-default-server)
	    (setq mud-default-server mud-default-default-server)))))

(defun mud-servers ()
  (mud-check-entry-file)
  mud-servers)

(defun mud-default-server ()
  (mud-check-entry-file)
  mud-default-server)

(defun mud-login-for-server (server)
  (mud-check-entry-file)
  (or (cdr (assoc server mud-logins))
      mud-default-login))

;;;; Commands and filters for specific MUDs:  TinyMUD.

;;; TinyMUD

(defvar tinymud-filter-hook
  '(mud-check-reconnect mud-check-page mud-fill-lines)
  "*List of functions to call on each line of tinymud output.  The
function is called with no arguments and the buffer narrowed to just
the line.") 

(defvar tinymud-connection-error-string
  "Either that player does not exist, or has a different password.")

(defvar tinymud-macro-commands-file "~/.tinymud_macros"
  "*Pathname of tinymud macros.")

(setq tinymud-output-filter nil)

;;; &&& WRONG!  Nobody ever bothered to write TINYMUD-FILTER.  The only reason that
;;; this ever worked is because OPEN-MUD always unconditionally replaced whatever
;;; filter was in effect immediately after a connection with MUD-FILTER.  So in the
;;; brief instant between when this function jams TINYMUD-FILTER in and when
;;; OPEN-MUD replaces it with MUD-FILTER, output from the server will cause us to
;;; bomb.  It seems that this was just unlikely enough not to come up.  This is
;;; probably related to the scoping problems because we're using symbol properties
;;; instead of defstruct, and makes me wonder about MUD-FILTER, because it tries to
;;; run other filters by calling RUN-HOOKS on MUD-FILTERS, and given that this
;;; apparently never blew out, perhaps MUD-FILTER was never being run with the
;;; proper environment, either, and hence it's slightly broken as well.  I'm not
;;; going to try to figure that out right now. For the moment, I've replaced
;;; it with MUD-FILTER, since that's what really happens anyway.  Either OPEN-MUD
;;; should not unconditionally jam a filter in, and should depend on the connection
;;; filter doing so, or this connection filter should not try to set a filter,
;;; given that OPEN-MUD just overrides it.  Either way, of course, if we try to
;;; actually use TINYMUD-FILTER, somebody should write it... --- Foner 26 Dec 92.
(defun tinymud-connect-filter (proc string)
  "Filter for connecting to a TinyMUD server.  Replaced with tinymud-filter
once successful."
  (if (not (string-equal string tinymud-connection-error-string))
;     (set-process-filter proc 'tinymud-filter)	; &&& WRONG!
      ;; &&& Still wrong, because OPEN-MUD overrides it, but it overrides it with the same thing.
      ;; We should get our act together on who's supposed to set the postconnection filter,
      ;; the connection filter itself or OPEN-MUD.
      (set-process-filter proc 'mud-filter)))

;;;; Comands and filters for specific MUDs:  SMUG (TinyMUD 2).

;;; Utility for SMUG-FILTER-HOOK.
(defun mud-convert-tabs-to-newlines ()
  "Replace all TABs to NEWLINEs in displaying of Smug output, since 
they represent new statements in the embedded programming language."
  (subst-char-in-region (point-min) (point-max) ?\t ?\n t))

(defvar smug-filter-hook
  '(mud-convert-tabs-to-newlines mud-fill-lines)
  "*List of functions to call on each line of Smug output.  The
function is called with no arguments and the buffer narrowed to just
the line.")

;;; Utility for SMUG-MACRO-COMMAND-FILTER-HOOK.
(defun mud-convert-newlines-to-tabs-in-strings ()
  "Replace all NEWLINEs present inside top-level strings with TABs, 
as they are likely code objects."
  (goto-char (point-min))
  (if (re-search-forward "[\\[\"]" (point-max) t)
      (progn 
	(forward-char -1)
	(subst-char-in-region (point)
			      (save-excursion (forward-sexp 1) (point))
			      ?\n ?\t t))))

;;; [Why is this a SETQ and not a DEFVAR?  --- Foner 26 Dec 92.]
(setq smug-macro-command-filter-hook
      '(mud-convert-newlines-to-tabs-in-strings))

(defvar smug-macro-commands-file "~/.smug_macros"
  "*Pathname of SMUG macros.")

;;;; Commands and filters for specific MUDs:  MOO.

(defvar moo-mode-hook '(define-moo-mode-commands))

(defun define-moo-mode-commands ()
  (define-key (current-local-map) "\^c\^d" 'moo-get-description)
  (define-key (current-local-map) "\^ch" 'moo-get-help)
  (define-key (current-local-map) "\^c\^f" 'moo-get-field)
  (define-key (current-local-map) "\^c\^v" 'moo-get-verb-listing))

(defun moo-explode-message ()
  "Convert a list of strings into more readable/editable text."
  (interactive)
  (goto-char (point-min))
  (while (search-forward "{\"" nil t)
    (replace-match "{\n"))
  (goto-char (point-min))
  (while (search-forward "\", \"" nil t)
    (replace-match "\n"))
  (goto-char (point-min))
  (while (search-forward "\"}" nil t)
    (replace-match "\n}"))
  (goto-char (1- (point-max)))
  (if (looking-at "\n")
      (delete-char 1)))

(defun moo-implode-message ()
  "Convert readable/editable text into a list of strings."
  (interactive)
  (goto-char (point-min))
  (while (search-forward "{\n" nil t)
    (replace-match "{\""))
  (goto-char (point-min))
  (while (search-forward "\n}" nil t)
    (replace-match "\"}"))
  (goto-char (point-min))
  (while (search-forward "\n" nil t)
    (replace-match "\", \"")))

(defun moo-get-help ()
  "Fetch a given help text and explode it."
  (interactive)
  (moo-do-fetch "Edit which help text: "
		"%s"
		"@show $help.%s"
		'moo-fix-help))

(defun moo-fix-help ()
  (save-excursion
    (end-of-line)
    (delete-region (point-min) (point))
    (insert "$help.(\"" moo-object "\")"))
  (moo-fix-field)
  (moo-explode-message))

(defun moo-get-field ()
  "Fetch the value of some field."
  (interactive)
  (moo-do-fetch "Edit what field: "
		"%s"
		"@show %s"
		'moo-fix-field))

(defun moo-get-description ()
  "Fetch the description of some object."
  (interactive)
  (moo-do-fetch "Edit description of what object: "
		"%s"
		"@show %s.description"
		'moo-fix-field))

(defun moo-fix-field ()
  (define-key (current-local-map) "\^c\^e" 'moo-explode-message)
  (define-key (current-local-map) "\^c\^i" 'moo-implode-message)
  (insert "; !(")
  (end-of-line)
  (insert " = ")
  (let ((start (point)))
    (re-search-forward "Value: *")
    (delete-region start (point)))
  (save-excursion
    (end-of-line)
    (insert ")")))

(defun moo-get-verb-listing ()
  "Fetch the MOO code for a particular verb."
  (interactive)
  (moo-do-fetch "Program what verb: "
		"%s"
		"@list %s without numbers"
		'moo-fix-listing))

(defun moo-fix-listing ()
  (cond ((looking-at "That object")
	 (let ((message (substring (buffer-string) 0 -1)))
	   (erase-buffer)
	   (error message)))
	((looking-at "That verb")
	 (let ((start (point)))
	   (end-of-line)
	   (delete-region start (point)))))
  (insert (concat "@program " moo-object "\n"))
  (if (looking-at "#")				; Kill the header line.
      (let ((start (point)))
	(beginning-of-line 2)
	(delete-region start (point))))
  (goto-char (point-max))
  (insert ".\n")
  (goto-char (point-min))
  (beginning-of-line 2))

(defun moo-do-fetch (prompt object-fmt command-fmt fixer)
  (setq moo-object (format object-fmt (read-string prompt)))
  (setq moo-state 'waiting)
  (setq moo-fixer fixer)
  (setq mud-current-process (get-buffer-process (current-buffer)))
  (pop-to-buffer (get-buffer-create moo-object))
  (erase-buffer)
  (mud-send-string (concat "PREFIX " moo-prefix
			   "\nSUFFIX " moo-suffix
			   "\n"
			   (format command-fmt moo-object)
			   "\nPREFIX\nSUFFIX\n")
		   mud-current-process))

(defvar moo-prefix "===MOO-Prefix===")
(defvar moo-suffix "===MOO-Suffix===")
(defvar moo-state 'idle)
(defvar moo-object nil)
(defvar moo-fixer nil)
(defvar moo-filter-hook
  '(moo-filter mud-check-page mud-check-reconnect mud-fill-lines))

(defun moo-filter ()
  (if (not (eq moo-state 'idle))
      (progn
	(goto-char (point-min))
	(while (not (eobp))
	  (let ((start (point)))
	    (cond ((and (eq moo-state 'waiting)
			(looking-at (concat moo-prefix "\n")))
		   (delete-region start (progn (beginning-of-line 2)
					       (point)))
		   (setq moo-state 'copying))
		  ((eq moo-state 'copying)
		   (cond ((looking-at moo-suffix)
			  (setq moo-state 'idle)
			  (beginning-of-line 2)
			  (delete-region start (point))
			  (let ((buff (current-buffer)))
			    (set-buffer (get-buffer moo-object))
			    (goto-char (point-min))
			    (mud-macro-expansion-mode)
			    (funcall moo-fixer)
			    (set-buffer buff)))
			 (t
			  (beginning-of-line 2)
			  (let* ((buff (current-buffer))
				 (str (buffer-substring start (point)))
				 (len (length str)))
			    (if (or (> len (length moo-suffix))
				    (not (equal (substring moo-suffix 0 len)
						str)))
				(progn
				  (delete-region start (point))
				  (set-buffer (get-buffer moo-object))
				  (goto-char (point-max))
				  (insert str)
				  (set-buffer buff)))))))
		  (t
		   (beginning-of-line 2))))))))

;;;; Utilities.

(defun mud-cleanup-extra-processes ()
  (interactive)
  (mapcar '(lambda (p)
	     (if (not (buffer-name (process-buffer (get-process p))))
		 (delete-process p)))
	  (process-list)))

;;;; End of file.
