
;;; wrapper for the "Pretty-Good-Privacy" program.
;;;

(provide 'pgp-lib)

(defvar pgp-program (expand-file-name "/afs/sipb.mit.edu/project/pgp/bin/pgp"))
(require 'terminal)
(require 'backquote)

(defvar pgp-encrypt-mail t)
(defvar pgp-sign-mail t)

;; a "continuation".. what we do after pgp is done..
;; this could set the current buffer (the terminal emulator one) 
;; into a new major mode (pgp-after-mode) to let you do things like:
;;  a) easily dismiss the pgp window
;;  b) view any output files.
;;  c) see what the "labelling" on the files was
;; (i.e., signature, encrypted); this shows up in mode line, not file!

(defun pgp-continue-frob (out-filename)
  (cond ((file-exists-p out-filename)
	 (set-buffer (get-buffer-create "*PGP Output*"))
	 (erase-buffer)
	 (insert-file-contents out-filename)
	 (delete-file out-filename)
	 (switch-to-buffer "*PGP Output*"))
	(t (message "PGP command completed with no output file" ))))

(defun pgp-continue-in-place  (out-file buf min max cont)
  (cond ((file-exists-p out-file)
	 (switch-to-buffer buf)
	 (goto-char min)
	 (delete-region min max)
	 (insert-file-contents out-file)
	 (delete-file out-file)
	 (apply cont nil))
	(t (message "PGP command completed with no output file"))))

(defun pgp-mail-continue ()
  (mail-to)
  (insert (save-excursion
	    (set-buffer "*PGP*")	;!!!
	    (goto-char (point-min))
	    (re-search-forward "Recipient's")
	    (re-search-forward "user ID: \\(.*$\\)")
	    (buffer-substring (match-beginning 1) (match-end 1)))))

(defun pgp-frob-region-1 (min max cont1 &optional opt1 opt2)
  (let ((temp-filename "/tmp/pgpxxx.txt")
	(out-filename "/tmp/pgpxxx.asc"))
    (if (file-exists-p temp-filename)
	(delete-file temp-filename))
    (if (file-exists-p out-filename)
	(delete-file out-filename))
    (write-region min max temp-filename)
    
    (let ((buf (get-buffer-create "*PGP*")))
      (switch-to-buffer buf)
      (erase-buffer)
      (let ((terminal-mode-hook
	     (function
	      (lambda ()
		(make-variable-buffer-local 'terminal-finished-hook) 
		(setq terminal-finished-hook (apply cont1 out-filename nil))))))
	(terminal-emulator buf
			   pgp-program
			   (nconc (list "-o" out-filename)
				  opt1
				  (list temp-filename)
				  opt2))))))

(defun pgp-frob-region  (min max &optional opt1 opt2)
  (pgp-frob-region-1 min max
		     (function
		      (lambda (out-filename)
			      (` (lambda () 
				   (pgp-continue-frob (, out-filename))))))
		     opt1 opt2))

(defun pgp-frob-region-in-place (min max &optional opt1 opt2 cont)
  (let ((cur-buf (current-buffer)))
    (pgp-frob-region-1 min max 
		       (function
			(lambda (out-filename)
			  (` (lambda () 
			       (pgp-continue-in-place (, out-filename)
						      (, cur-buf)
						      (, min)
						      (, max)
						      (quote (, cont)))))))
		       opt1 opt2)))

(defun pgp-encrypt-ascii-region (min max to)
  (interactive "r\nsRecipient name: ")
  (pgp-frob-region min max (list "-eaw") (list to)))

(defun pgp-decrypt-ascii-region (min max)
  (interactive "r")
  (pgp-frob-region min max nil nil))

(defun pgp-sign-ascii-region (min max)
  (interactive "r")
  (pgp-frob-region min max (list "-saw") nil))

(defun pgp-encrypt-ascii-buffer (to)
  (interactive "sRecipient name: ")
  (pgp-encrypt-ascii-region (point-min) (point-max) (list to)))

(defun pgp-decrypt-ascii-buffer ()
  (interactive)
  (pgp-decrypt-ascii-region (point-min) (point-max)))

(defun pgp-sign-encrypt-ascii-buffer (to)
  (interactive "sRecipient name: ")
  (pgp-frob-region (point-min) (point-max) (list "-seaw") (list to)))

(defun pgp-sign-encrypt-ascii-buffer-in-place (to)
  (interactive "s(in place) Recipient name: ")
  (pgp-frob-region-in-place (point-min) (point-max) (list "-seaw") (list to)))

(defun pgp-frob-mail (to)
  (interactive "sMail Recipient name: ")
  (goto-char (point-min))
  (re-search-forward mail-header-separator)
  (goto-char (match-beginning 0))
  (forward-line 1)
  (save-excursion
    (insert (buffer-substring (point-min) (point))))
  (pgp-frob-region-in-place (point) (point-max) (list "-eaw") (list to)
			      'pgp-mail-continue))

(global-set-key "\C-c\C-p\C-e" 'pgp-encrypt-ascii-buffer)
(global-set-key "\C-c\C-p\C-s" 'pgp-sign-ascii-buffer)
(global-set-key "\C-c\C-p\C-m" 'pgp-frob-mail)
(global-set-key "\C-c\C-p\C-d" 'pgp-decrypt-ascii-buffer)

(global-set-key "\C-cpm" 'pgp-frob-mail)

(defun te-sentinel (process message)
  (cond ((eq (process-status process) 'run))
	((null (buffer-name (process-buffer process)))) ;deleted
	(t (let ((b (current-buffer))
		 (done-hooks (and
			      (boundp 'terminal-finished-hook)
			      terminal-finished-hook)))

	     (save-excursion
	       (set-buffer (process-buffer process))
	       (setq buffer-read-only nil)
	       (fundamental-mode)
	       (goto-char (point-max))
	       (delete-blank-lines)
	       (delete-horizontal-space)
	       (insert "\n*******\n" message "*******\n"))
	     (if (and (eq b (process-buffer process))
		      (waiting-for-user-input-p))
		 (progn (goto-char (point-max))
			(recenter -1)))
	     (run-hooks 'done-hooks)))))
